# This file is part of the DiscoPoP software (http://www.discopop.tu-darmstadt.de)
#
# Copyright (c) 2020, Technische Universitaet Darmstadt, Germany
#
# This software may be modified and distributed under the terms of
# the 3-Clause BSD License.  See the LICENSE file in the package base
# directory for details.
from typing import List


class Arguments(object):
    no_gui: bool = False
    execute_all_configurations: bool = False
    execute_configurations_with_tag: List[str] = []
